@@ -258,6 +258,15 @@ class Puppet::Node::Environment
     !original_manifest.nil? && !original_manifest.empty? && original_manifest != Puppet[:default_manifest]
   end
 
+  # @api private
+  def future_parser?
+    if @future_parser.nil?
+      environment_conf = Puppet.lookup(:environments).get_conf(name)
+      @future_parser = 'future' == (environment_conf.nil? ? Puppet[:parser] : environment_conf.parser)
+    end
+    @future_parser
+  end
+
   # Return an environment-specific Puppet setting.
   #
   # @api public
@@ -550,7 +559,7 @@ class Puppet::Node::Environment
       if file == NO_MANIFEST
         Puppet::Parser::AST::Hostclass.new('')
       elsif File.directory?(file)
-        if Puppet[:parser] == 'future'
+        if Puppet.future_parser?
           parse_results = Puppet::FileSystem::PathPattern.absolute(File.join(file, '**/*.pp')).glob.sort.map do | file_to_parse |
             parser.file = file_to_parse
             parser.parse
@@ -568,6 +577,10 @@ class Puppet::Node::Environment
         parser.parse
       end
     end
+  rescue Puppet::ParseErrorWithIssue => detail
+    @known_resource_types.parse_failed = true
+    detail.environment = self.name
+    raise
   rescue => detail
     @known_resource_types.parse_failed = true
 
