@@ -197,7 +197,7 @@ class Puppet::Pops::Types::TypeCalculator
     data_variant.addTypes(@data_hash.copy)
     data_variant.addTypes(@data_array.copy)
     data_variant.addTypes(Types::PScalarType.new)
-    data_variant.addTypes(Types::PNilType.new)
+    data_variant.addTypes(Types::PUndefType.new)
     data_variant.addTypes(@data_tuple_t.copy)
     @data_variant_t = data_variant
 
@@ -212,7 +212,7 @@ class Puppet::Pops::Types::TypeCalculator
     non_empty_string.size_type.to = nil # infinity
     @non_empty_string_t = non_empty_string
 
-    @nil_t = Types::PNilType.new
+    @nil_t = Types::PUndefType.new
   end
 
   # Convenience method to get a data type for comparisons
@@ -269,14 +269,21 @@ class Puppet::Pops::Types::TypeCalculator
     if t2.is_a?(Class)
       t2 = type(t2)
     end
-    # Unit can be assigned to anything
-    return true if t2.class == Types::PUnitType
+    t2_class = t2.class
 
-    if t2.class == Types::PVariantType
+    # Unit can be assigned to anything
+    return true if t2_class == Types::PUnitType
+
+    if t2_class == Types::PVariantType
       # Assignable if all contained types are assignable
       t2.types.all? { |vt| @@assignable_visitor.visit_this_1(self, t, vt) }
     else
-      @@assignable_visitor.visit_this_1(self, t, t2)
+      # Turn NotUndef[T] into T when T is not assignable from Undef
+      if t2_class == Types::PNotUndefType && !(t2.type.nil? || assignable?(t2.type, @nil_t))
+        assignable?(t, t2.type)
+      else
+        @@assignable_visitor.visit_this_1(self, t, t2)
+      end
     end
  end
 
@@ -322,7 +329,7 @@ class Puppet::Pops::Types::TypeCalculator
     when c == Regexp
       type = Types::PRegexpType.new()
     when c == NilClass
-      type = Types::PNilType.new()
+      type = Types::PUndefType.new()
     when c == FalseClass, c == TrueClass
       type = Types::PBooleanType.new()
     when c == Class
@@ -354,10 +361,23 @@ class Puppet::Pops::Types::TypeCalculator
     # do nothing, there is nothing to change for most types
   end
 
+  # @return [Boolean] true if the given argument is contained in a struct element key
+  def is_struct_element_key?(o)
+    c = o.eContainer
+    if c.is_a?(Types::POptionalType)
+      o = c
+      c = c.eContainer
+    end
+    c.is_a?(Types::PStructElement) && c.key_type.equal?(o)
+  end
+  private :is_struct_element_key?
+
   def generalize_PStringType(o)
-    o.values = []
-    o.size_type = nil
-    []
+    # Skip generalization if the string is contained in a PStructElement key.
+    unless is_struct_element_key?(o)
+      o.values = []
+      o.size_type = nil
+    end
   end
 
   def generalize_PCollectionType(o)
@@ -430,6 +450,19 @@ class Puppet::Pops::Types::TypeCalculator
     return x < y ? x <= o && y >= o : y <= o && x >= o
   end
 
+  # @api private
+  def instance_of_PStringType(t, o)
+    return false unless o.is_a?(String)
+    # true if size compliant
+    size_t = t.size_type
+    if size_t.nil? || instance_of_PIntegerType(size_t, o.size)
+      values = t.values
+      values.empty? || values.include?(o)
+    else
+      false
+    end
+  end
+
   def instance_of_PTupleType(t, o)
     return false unless o.is_a?(Array)
     # compute the tuple's min/max size, and check if that size matches
@@ -444,9 +477,18 @@ class Puppet::Pops::Types::TypeCalculator
 
   def instance_of_PStructType(t, o)
     return false unless o.is_a?(Hash)
-    h = t.hashed_elements
-    # all keys must be present and have a value (even if nil/undef)
-    (o.keys - h.keys).empty? && h.all? { |k,v| instance_of(v, o[k]) }
+    matched = 0
+    t.elements.all? do |e|
+      key = e.name
+      v = o[key]
+      if v.nil? && !o.include?(key)
+        # Entry is missing. Only OK when key is optional
+        assignable?(e.key_type, @nil_t)
+      else
+        matched += 1
+        instance_of(e.value_type, v)
+      end
+    end && matched == o.size
   end
 
   def instance_of_PHashType(t, o)
@@ -463,12 +505,16 @@ class Puppet::Pops::Types::TypeCalculator
     instance_of(@data_variant_t, o)
   end
 
-  def instance_of_PNilType(t, o)
+  def instance_of_PNotUndefType(t, o)
+    !(o.nil? || o == :undef) && (t.type.nil? || instance_of(t.type, o))
+  end
+
+  def instance_of_PUndefType(t, o)
     o.nil? || o == :undef
   end
 
   def instance_of_POptionalType(t, o)
-    instance_of_PNilType(t, o) || instance_of(t.optional_type, o)
+    instance_of_PUndefType(t, o) || instance_of(t.optional_type, o)
   end
 
   def instance_of_PVariantType(t, o)
@@ -497,11 +543,11 @@ class Puppet::Pops::Types::TypeCalculator
     return t.is_a?(Types::PAnyType)
   end
 
-  # Answers if t represents the puppet type PNilType
+  # Answers if t represents the puppet type PUndefType
   # @api public
   #
   def is_pnil?(t)
-    return t.nil? || t.is_a?(Types::PNilType)
+    return t.nil? || t.is_a?(Types::PUndefType)
   end
 
   # Answers, 'What is the common type of t1 and t2?'
@@ -590,7 +636,8 @@ class Puppet::Pops::Types::TypeCalculator
 
     if t1.is_a?(Types::PStringType) && t2.is_a?(Types::PStringType)
       t = Types::PStringType.new()
-      t.values = t1.values | t2.values
+      t.values = t1.values | t2.values unless t1.values.empty? || t2.values.empty?
+      t.size_type = common_type(t1.size_type, t2.size_type) unless t1.size_type.nil? || t2.size_type.nil?
       return t
     end
 
@@ -787,7 +834,50 @@ class Puppet::Pops::Types::TypeCalculator
 
   # @api private
   def infer_NilClass(o)
-    Types::PNilType.new()
+    Types::PUndefType.new()
+  end
+
+  # @api private
+  # @param o [Proc]
+  def infer_Proc(o)
+    min = 0
+    max = 0
+    if o.respond_to?(:parameters)
+      mapped_types = o.parameters.map do |p|
+        param_t = Types::PAnyType.new
+        case p[0]
+        when :rest
+          max = :default
+          break param_t
+        when :req
+          min += 1
+        end
+        max += 1
+      	param_t
+      end
+    else
+      # Cannot correctly compute the signature in Ruby 1.8.7 because arity for
+      # optional values is screwed up (there is no way to get the upper limit),
+      # an optional looks the same as a varargs.
+      arity = o.arity
+      if arity < 0
+        min = -arity - 1
+        max = :default # i.e. infinite (which is wrong when there are optional - flaw in 1.8.7)
+      else
+        min = max = arity
+      end
+      mapped_types = Array.new(min) { Types::PAnyType.new }
+    end
+    if min == 0 || min != max
+      mapped_types << min
+      mapped_types << max
+    end
+    Types::TypeFactory.callable(*mapped_types)
+  end
+
+  # @api private
+  def infer_PuppetProc(o)
+    infer_Closure(o.closure)
   end
 
   # Inference of :default as PDefaultType, and all other are Ruby[Symbol]
@@ -796,6 +886,8 @@ class Puppet::Pops::Types::TypeCalculator
     case o
     when :default
       Types::PDefaultType.new()
+    when :undef
+      Types::PUndefType.new()
     else
       infer_Object(o)
     end
@@ -831,7 +923,7 @@ class Puppet::Pops::Types::TypeCalculator
     type = Types::PArrayType.new()
     type.element_type =
     if o.empty?
-      Types::PNilType.new()
+      Types::PUndefType.new()
     else
       infer_and_reduce_type(o)
     end
@@ -843,8 +935,8 @@ class Puppet::Pops::Types::TypeCalculator
   def infer_Hash(o)
     type = Types::PHashType.new()
     if o.empty?
-      ktype = Types::PNilType.new()
-      etype = Types::PNilType.new()
+      ktype = Types::PUndefType.new()
+      etype = Types::PUndefType.new()
     else
       ktype = infer_and_reduce_type(o.keys())
       etype = infer_and_reduce_type(o.values())
@@ -871,7 +963,7 @@ class Puppet::Pops::Types::TypeCalculator
   def infer_set_Array(o)
     if o.empty?
       type = Types::PArrayType.new()
-      type.element_type = Types::PNilType.new()
+      type.element_type = Types::PUndefType.new()
       type.size_type = size_as_type(o)
     else
       type = Types::PTupleType.new()
@@ -881,19 +973,29 @@ class Puppet::Pops::Types::TypeCalculator
   end
 
   def infer_set_Hash(o)
-    type = Types::PHashType.new()
     if o.empty?
-      ktype = Types::PNilType.new()
-      vtype = Types::PNilType.new()
+      type = Types::PHashType.new
+      type.key_type = Types::PUndefType.new
+      type.element_type = Types::PUndefType.new
+      type.size_type = size_as_type(o)
+    elsif o.keys.all? {|k| instance_of_PStringType(@non_empty_string_t, k) }
+      type = Types::PStructType.new
+      type.elements = o.map do |k,v|
+        element = Types::PStructElement.new
+        element.key_type = infer_String(k)
+        element.value_type = infer_set(v)
+        element
+      end
     else
-      ktype = Types::PVariantType.new()
-      ktype.types = o.keys.map() {|k| infer_set(k) }
-      etype = Types::PVariantType.new()
-      etype.types = o.values.map() {|e| infer_set(e) }
+      type = Types::PHashType.new
+      ktype = Types::PVariantType.new
+      ktype.types = o.keys.map {|k| infer_set(k) }
+      etype = Types::PVariantType.new
+      etype.types = o.values.map {|e| infer_set(e) }
+      type.key_type = unwrap_single_variant(ktype)
+      type.element_type = unwrap_single_variant(etype)
+      type.size_type = size_as_type(o)
     end
-    type.key_type = unwrap_single_variant(ktype)
-    type.element_type = unwrap_single_variant(etype)
-    type.size_type = size_as_type(o)
     type
   end
 
@@ -917,9 +1019,14 @@ class Puppet::Pops::Types::TypeCalculator
   end
 
   # @api private
-  def assignable_PNilType(t, t2)
+  def assignable_PNotUndefType(t, t2)
+    !assignable?(t2, @nil_t) && (t.type.nil? || assignable?(t.type, t2))
+  end
+
+  # @api private
+  def assignable_PUndefType(t, t2)
     # Only undef/nil is assignable to nil type
-    t2.is_a?(Types::PNilType)
+    t2.is_a?(Types::PUndefType)
   end
 
   # Anything is assignable to a Unit type
@@ -964,11 +1071,7 @@ class Puppet::Pops::Types::TypeCalculator
     to = range.to
     x = from.nil? ? 1 : from
     y = to.nil? ? TheInfinity : to
-    if x < y
-      [x, y]
-    else
-      [y, x]
-    end
+    [x, y]
   end
 
   # @api private
@@ -1013,7 +1116,7 @@ class Puppet::Pops::Types::TypeCalculator
     if args_tuple.size_type
       raise ArgumentError, "Callable tuple may not have a size constraint when used as args"
     end
-    # Assume no block was given - i.e. it is nil, and its type is PNilType
+    # Assume no block was given - i.e. it is nil, and its type is PUndefType
     block_t = @nil_t
     if self.class.is_kind_of_callable?(args_tuple.types.last)
       # a split is needed to make it possible to use required, optional, and varargs semantics
@@ -1045,6 +1148,17 @@ class Puppet::Pops::Types::TypeCalculator
     end
   end
 
+  # @api private
+  def self.is_kind_of_optional?(t, optional = true)
+    case t
+    when Types::POptionalType
+      true
+    when Types::PVariantType
+      t.types.all? {|t2| is_kind_of_optional?(t2, optional) }
+    else
+      false
+    end
+  end
 
   def callable_PArrayType(args_array, callable_t)
     return false unless assignable?(callable_t.param_types, args_array)
@@ -1052,8 +1166,8 @@ class Puppet::Pops::Types::TypeCalculator
     assignable?(callable_t.block_type || @nil_t, @nil_t)
   end
 
-  def callable_PNilType(nil_t, callable_t)
-    # if callable_t is Optional (or indeed PNilType), this means that 'missing callable' is accepted
+  def callable_PUndefType(nil_t, callable_t)
+    # if callable_t is Optional (or indeed PUndefType), this means that 'missing callable' is accepted
     assignable?(callable_t, nil_t)
   end
 
@@ -1127,22 +1241,35 @@ class Puppet::Pops::Types::TypeCalculator
   # @api private
   #
   def assignable_PStructType(t, t2)
-    return true if t == t2 || t.elements.empty? && (t2.is_a?(Types::PHashType))
-    h = t.hashed_elements
     if t2.is_a?(Types::PStructType)
       h2 = t2.hashed_elements
-      h.size == h2.size && h.all? {|k, v| assignable?(v, h2[k]) }
+      matched = 0
+      t.elements.all? do |e1|
+        e2 = h2[e1.name]
+        if e2.nil?
+          assignable?(e1.key_type, @nil_t)
+        else
+          matched += 1
+          assignable?(e1.key_type, e2.key_type) && assignable?(e1.value_type, e2.value_type)
+        end
+      end && matched == h2.size
     elsif t2.is_a?(Types::PHashType)
-      size_t2 = t2.size_type || @collection_default_size_t
-      size_t = Types::PIntegerType.new
-      size_t.from = size_t.to = h.size
-      # compatible size
-      # hash key type must be string of min 1 size
-      # hash value t must be assignable to each key
-      element_type = t2.element_type
-      assignable_PIntegerType(size_t, size_t2) &&
-        assignable?(@non_empty_string_t, t2.key_type) &&
-        h.all? {|k,v| assignable?(v, element_type) }
+      required = 0
+      required_elements_assignable = t.elements.all? do |e|
+        if assignable?(e.key_type, @nil_t)
+          true
+        else
+          required += 1
+          assignable?(e.value_type, t2.element_type)
+        end
+      end
+      if required_elements_assignable
+        size_t2 = t2.size_type || @collection_default_size_t
+        size_t = Types::PIntegerType.new
+        size_t.from = required
+        size_t.to = t.elements.size
+        assignable_PIntegerType(size_t, size_t2)
+      end
     else
       false
     end
@@ -1150,9 +1277,10 @@ class Puppet::Pops::Types::TypeCalculator
 
   # @api private
   def assignable_POptionalType(t, t2)
-    return true if t2.is_a?(Types::PNilType)
+    return true if t2.is_a?(Types::PUndefType)
+    return true if t.optional_type.nil?
     if t2.is_a?(Types::POptionalType)
-      assignable?(t.optional_type, t2.optional_type)
+      assignable?(t.optional_type, t2.optional_type || @t)
     else
       assignable?(t.optional_type, t2)
     end
@@ -1323,11 +1451,11 @@ class Puppet::Pops::Types::TypeCalculator
   # @api private
   def assignable_PArrayType(t, t2)
     if t2.is_a?(Types::PArrayType)
-      return false unless assignable?(t.element_type, t2.element_type)
+      return false unless t.element_type.nil? || assignable?(t.element_type, t2.element_type || @t)
       assignable_PCollectionType(t, t2)
 
     elsif t2.is_a?(Types::PTupleType)
-      return false unless t2.types.all? {|t2_element| assignable?(t.element_type, t2_element) }
+      return false unless t.element_type.nil? || t2.types.all? {|t2_element| assignable?(t.element_type, t2_element) }
       t2_regular = t2.types[0..-2]
       t2_ranged = t2.types[-1]
       t2_from, t2_to = size_range(t2.size_type)
@@ -1364,7 +1492,9 @@ class Puppet::Pops::Types::TypeCalculator
   def assignable_PHashType(t, t2)
     case t2
     when Types::PHashType
-      return false unless assignable?(t.key_type, t2.key_type) && assignable?(t.element_type, t2.element_type)
+      return true if (t.size_type.nil? || t.size_type.from == 0) && t2.is_the_empty_hash?
+      return false unless t.key_type.nil? || assignable?(t.key_type, t2.key_type || @t)
+      return false unless t.element_type.nil? || assignable?(t.element_type, t2.element_type || @t)
       assignable_PCollectionType(t, t2)
     when Types::PStructType
       # hash must accept String as key type
@@ -1373,10 +1503,10 @@ class Puppet::Pops::Types::TypeCalculator
       size_t = t.size_type || @collection_default_size_t
       min, max = size_t.range
       struct_size = t2.elements.size
+      key_type = t.key_type
       element_type = t.element_type
       ( struct_size >= min && struct_size <= max &&
-        assignable?(t.key_type, @non_empty_string_t)  &&
-        t2.hashed_elements.all? {|k,v| assignable?(element_type, v) })
+        t2.elements.all? {|e| (key_type.nil? || instance_of(key_type, e.name)) && (element_type.nil? || assignable?(element_type, e.value_type)) })
     else
       false
     end
@@ -1408,7 +1538,15 @@ class Puppet::Pops::Types::TypeCalculator
   # Data is assignable by other Data and by Array[Data] and Hash[Scalar, Data]
   # @api private
   def assignable_PDataType(t, t2)
-    t2.is_a?(Types::PDataType) || assignable?(@data_variant_t, t2)
+    # We cannot put the NotUndefType[Data] in the @data_variant_t since that causes an endless recursion
+    case t2
+    when Types::PDataType
+      true
+    when Types::PNotUndefType
+      assignable?(t, t2.type || @t)
+    else
+      assignable?(@data_variant_t, t2)
+    end
   end
 
   # Assignable if t2's has the same runtime and the runtime name resolves to
@@ -1453,7 +1591,7 @@ class Puppet::Pops::Types::TypeCalculator
   def string_PAnyType(t)     ; "Any"     ; end
 
   # @api private
-  def string_PNilType(t)     ; 'Undef'   ; end
+  def string_PUndefType(t)     ; 'Undef'   ; end
 
   # @api private
   def string_PDefaultType(t) ; 'Default' ; end
@@ -1579,7 +1717,16 @@ class Puppet::Pops::Types::TypeCalculator
   end
 
   def string_PStructElement(t)
-    "'#{t.name}'=>#{string(t.type)}"
+    k = t.key_type
+    value_optional = assignable?(t.value_type, @nil_t)
+    key_string =
+      if k.is_a?(Types::POptionalType)
+        # Output as literal String
+        value_optional ? "'#{t.name}'" : string(k)
+      else
+        value_optional ? "NotUndef['#{t.name}']" : "'#{t.name}'"
+      end
+    "#{key_string}=>#{string(t.value_type)}"
   end
 
   # @api private
@@ -1645,11 +1792,30 @@ class Puppet::Pops::Types::TypeCalculator
     end
   end
 
+  # @api private
+  def string_PNotUndefType(t)
+    contained_type = t.type
+    if contained_type.nil? || contained_type.class == Puppet::Pops::Types::PAnyType
+      'NotUndef'
+    else
+      if contained_type.is_a?(Puppet::Pops::Types::PStringType) && contained_type.values.size == 1
+        "NotUndef['#{contained_type.values[0]}']"
+      else
+        "NotUndef[#{string(contained_type)}]"
+      end
+    end
+  end
+
   def string_POptionalType(t)
-    if t.optional_type.nil?
+    optional_type = t.optional_type
+    if optional_type.nil?
       "Optional"
     else
-      "Optional[#{string(t.optional_type)}]"
+      if optional_type.is_a?(Puppet::Pops::Types::PStringType) && optional_type.values.size == 1
+        "Optional['#{optional_type.values[0]}']"
+      else
+        "Optional[#{string(optional_type)}]"
+      end
     end
   end
 
